@@ -433,10 +433,12 @@ void read_super_blocks(char* device, file_system_info* fs_info)
 
     fs_info->block_size  = btrfs_super_nodesize(root->fs_info->super_copy);
     fs_info->usedblocks  = btrfs_super_bytes_used(root->fs_info->super_copy) / fs_info->block_size;
+    fs_info->superBlockUsedBlocks = fs_info->usedblocks;
     fs_info->device_size = btrfs_super_total_bytes(root->fs_info->super_copy);
     fs_info->totalblock  = fs_info->device_size / fs_info->block_size;
     log_mesg(0, 0, 0, fs_opt.debug, "block_size = %i\n", fs_info->block_size);
     log_mesg(0, 0, 0, fs_opt.debug, "usedblock = %lli\n", fs_info->usedblocks);
+    log_mesg(0, 0, 0, fs_opt.debug, "superBlockUsedBlocks = %lli\n", fs_info->superBlockUsedBlocks);
     log_mesg(0, 0, 0, fs_opt.debug, "device_size = %llu\n", fs_info->device_size);
     log_mesg(0, 0, 0, fs_opt.debug, "totalblock = %lli\n", fs_info->totalblock);
 
